//Spin orbital interaction Buzdin model 21-11-2016
//For the fixed value of Josephsn frequency
//Signal is added to the voltage as gauss distribution with integral

// Pavlina's changes:
//18.03.2018:
// - stop the while cycle of t with while(++t<=tN+1); - that way is adding also the last point
// - all physical and numerical parameters are taken above before the main
// - define functions fmx, fmy, fmy and check the correcteness of The first coefficients of Runge-Kutta method
//19.03.2018:
// - define function fph,
// - check the correcteness of The second, third and fourth coefficients of Runge-Kutta method
// - v2: clean the code
// - v3: change the method from Runge-Kutta4 to Gauss-Legendre4
//20.03.2018:
// - changing in the begginig of the time-cycle -  if((t*TP>=(t1-(TP/2.)))&&(t*TP<=(t2-(TP/2.)))) (adding equality)

#include<iostream>
#include<cmath>
#include<iomanip>
#include<stdio.h>
#include<stdlib.h>
#include <stddef.h>
#include <omp.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <sstream>
#define MAXLEN 80

typedef struct point_s {
        double G;
        double r;
		double m;
} point;

struct sample_parameters
{
  char i_OmegaF[MAXLEN];
  char i_r_min[MAXLEN];
  char i_r_max[MAXLEN];
  char i_r_step[MAXLEN];
  char i_G_min[MAXLEN];
  char i_G_max[MAXLEN];
  char i_G_step[MAXLEN];
}
  sample_parameters;
  
 

/*
 * initialize data to default values
 */
void
init_parameters (struct sample_parameters * parms)
{
  strncpy (parms->i_OmegaF, "-25", MAXLEN);
  strncpy (parms->i_r_min, "-25", MAXLEN);
  strncpy (parms->i_r_max, "-25", MAXLEN);
  strncpy (parms->i_r_step, "-25", MAXLEN);
  strncpy (parms->i_G_min, "-25", MAXLEN);
  strncpy (parms->i_G_max, "-25", MAXLEN);
  strncpy (parms->i_G_step, "-25", MAXLEN);
 
}

/*
 * trim: get rid of trailing and leading whitespace...
 *       ...including the annoying "\n" from fgets()
 */
char *
trim (char * s)
{
  /* Initialize start, end pointers */
  char *s1 = s, *s2 = &s[strlen (s) - 1];

  /* Trim and delimit right side */
  while ( (isspace (*s2)) && (s2 >= s1) )
    s2--;
  *(s2+1) = '\0';

  /* Trim left side */
  while ( (isspace (*s1)) && (s1 < s2) )
    s1++;

  /* Copy finished string */
  strcpy (s, s1);
  return s;
}

/*
 * parse external parameters file
 *
 * NOTES:
 * - There are millions of ways to do this, depending on your
 *   specific needs.
 *
 * - In general:
 *   a) The client will know which parameters it's expecting
 *      (hence the "struct", with a specific set of parameters).
 *   b) The client should NOT know any specifics about the
 *      configuration file itself (for example, the client
 *      shouldn't know or care about it's name, its location,
 *      its format ... or whether or not the "configuration
 *      file" is even a file ... or a database ... or something
 *      else entirely).
 *   c) The client should initialize the parameters to reasonable
 *      defaults
 *   d) The client is responsible for validating whether the
 *      pararmeters are complete, or correct.
 */
void
parse_config (struct sample_parameters * parms, char * fl)
{
  char *s, buff[256];
  FILE *fp = fopen (fl, "r");
  if (fp == NULL)
  {
    return;
  }

  /* Read next line */
  while ((s = fgets (buff, sizeof buff, fp)) != NULL)
  {
    /* Skip blank lines and comments */
    if (buff[0] == '\n' || buff[0] == '#')
      continue;

    /* Parse name/value pair from line */
    char name[MAXLEN], value[MAXLEN];
    s = strtok (buff, "=");
    if (s==NULL)
      continue;
    else
      strncpy (name, s, MAXLEN);
    s = strtok (NULL, "=");
    if (s==NULL)
      continue;
    else
      strncpy (value, s, MAXLEN);
    trim (value);

    /* Copy into correct entry in parameters struct */
    if (strcmp(name, "OmegaF")==0)
      strncpy (parms->i_OmegaF, value, MAXLEN);
    else if (strcmp(name, "r_min")==0)
      strncpy (parms->i_r_min, value, MAXLEN);
    else if (strcmp(name, "r_max")==0)
      strncpy (parms->i_r_max, value, MAXLEN);
  else if (strcmp(name, "r_step")==0)
      strncpy (parms->i_r_step, value, MAXLEN);
  else if (strcmp(name, "G_min")==0)
      strncpy (parms->i_G_min, value, MAXLEN);
  else if (strcmp(name, "G_max")==0)
      strncpy (parms->i_G_max, value, MAXLEN);
  else if (strcmp(name, "G_step")==0)
      strncpy (parms->i_G_step, value, MAXLEN);
    else
      printf ("WARNING: %s/%s: Unknown name/value pair!\n",
        name, value);
  }

  /* Close file */
  fclose (fp);
}


//######################################################################################
// declearing all physical and numerical parameters
double
		Tmax=1000,//Time domain for one step of current
		TP=0.01,//Step of time
		Omega=0,//Josephson frequency
		OmegaF=1,
		
		
		alpha=0.5,//Dissipation parameter
		
		G_min=0,
		G_max=400,//200
		G_step=1,
		G,	//
		
		r,
		r_min=0,
		r_max=1,
		r_step=0.01,//
        //signal parameters
		As=1.5,// 
		dt=6,// 
		t0=25,// 
		V0=1, // = 1/w
		M=1;//Absolute value of the magnetic momentum

//######################################################################################
// declearing the functions of the mathematical model

double fmx (double t, double  mx, double my, double mz, double ph, double G, double r, double alpha)
// right-hand-side of dmx/dt = RHS equation
{
  double res;
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				mz*my-G*r*mz*sin(ph-r*my)
				+alpha*(mx*pow(mz,2)+G*r*mx*my*sin(ph-r*my))
			);
  return  res;
}
double fmy (double t, double  mx, double my, double mz, double ph, double G, double r, double alpha)
// right-hand-side of dmy/dt = RHS equation
{
  double res;
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				-1.*mz*mx
				+alpha*(my*pow(mz,2)-G*r*(pow(mz,2)+pow(mx,2))*sin(ph-r*my))
			);
  return  res;
}

double fmz (double t, double  mx, double my, double mz, double ph, double G, double r, double alpha)
// right-hand-side of dmz/dt = RHS equation
{
  double res;
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				G*r*mx*sin(ph-r*my)
				+alpha*(G*r*my*mz*sin(ph-r*my)-mz*(pow(mx,2)+pow(my,2)))
			);
  return  res;
}

double fph (double t, double  mx, double my, double mz, double ph, double Signal)
// right-hand-side of dph/dt = RHS equation
{
  double res;
    res=V0*(Signal-sin(ph-r*my));
  return  res;
}

int main(int argc, char* argv[])
	{
	using namespace std;
	
	FILE *stream;
	char mys[255];
	char str[50];
	
	struct sample_parameters parms;
	char * fl=argv[1];
	
	printf ("Initializing parameters to default values...\n");
	init_parameters (&parms);
	
	printf ("Reading config file...\n");
	parse_config (&parms,fl);
	
	OmegaF=atof(parms.i_OmegaF);

	r_min=atof(parms.i_r_min);
	r_max=atof(parms.i_r_max);
	r_step=atof(parms.i_r_step);
	
	G_min=atof(parms.i_G_min);
	G_max=atof(parms.i_G_max);
	G_step=atof(parms.i_G_step);
	
	int
		t,	//index of time
		tN;	//total number of time cycles
//######################################################################################
	double const
		pi=3.14159265358979323846;
    //preparing coefficients for the implicit RK => Gauss-legendre method
    double const
        c1 = 1/2. - (1/6.)*sqrt(3.);
    double const
        c2 = 1/2. + (1/6.)*sqrt(3.);
    double const
        a11 = 1/4.;
    double const
        a12 = 1/4. - (1/6.)*sqrt(3.);
    double const
        a21 = 1/4. + (1/6.)*sqrt(3.);
    double const
        a22 = 1/4.;

	double
		//~~~~~~~~~~~~~~~~~~~
		Ej=1,//Josephson energy
		Es,//Superconducting part of Energy
		Signal,
		Is,
		time,//time value for writing
		m,//Absolute value of the magnetic momentum for writing
        t1,t2,
		ph,//phase difference of JJ
		P1,P2,P3,P4,
		//components of magnetic momentum
		mx,	 mx1,	mx2,	mx3,	mx4,
		my,	 my1,	my2,	my3,	my4,
		mz,	 mz1,	mz2,	mz3,	mz4,

		mx1s, my1s, mz1s,
		mx2s, my2s, mz2s,
		P1s, P2s, s,
		mx1s1, my1s1, mz1s1,
		mx2s1, my2s1, mz2s1,
		P1s1, P2s1;

		t1=t0-(dt/2.);
		t2=t0+(dt/2.);
//#######################################################################################
	
	
	//Creation of streams for the result's files
	/*FILE
		*fmxtime,	*fmytime,	*fmztime,   *fmtime,
		*fsignal,   *fphtime,
		*fmzmx,		*fmzmy,     *fmymx,
		*fes,       *fIs,
		*fdebug;
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//Opening files
	fmxtime=fopen("mx_time.dat","w");
	fmytime=fopen("my_time.dat","w");
	fmztime=fopen("mz_time.dat","w");
	fsignal=fopen("Signal_time.dat","w");
	fphtime=fopen("ph_time.dat","w");
	fmtime=fopen("m_time.dat","w");
	fmzmx=fopen("mz_mx.dat","w");
	fmzmy=fopen("mz_my.dat","w");
	fmymx=fopen("my_mx.dat","w");
	fIs=fopen("Is_time.dat","w");
	fdebug = fopen("debug.dat","w");*/
	//*****************************************************************
	tN=(Tmax/TP);//number of points in time
	
	double start=0.0, stop=0.0;
	int rank, iterations=0.0;
	
	int i,j,k_mz_p1=0,k_mz_m1=0,k_mz_my_m1=0;
	int G_iter=(G_max-G_min)/G_step, r_iter=(r_max-r_min)/r_step;
	int all_iter=G_iter*r_iter;
	
	point point_arr_mz_p1[all_iter],point_arr_mz_m1[all_iter],point_arr_mz_my_m1[all_iter];
	
	//begining of parallel region
	start=omp_get_wtime();
	#pragma omp parallel private(i,j,G,r,rank,mx, my, mz, ph, t, Signal, mx1s, my1s, mz1s, P1s, mx2s, my2s, mz2s, P2s, s, mx1s1, my1s1, mz1s1, mx2s1, my2s1, mz2s1, P1s1, P2s1, m, Is, time)
	{
		rank=omp_get_thread_num();
		#pragma omp for
		for(i=0;i<G_iter;i++){
			G=G_min+(i*G_step);
			for(j=0;j<r_iter;j++){
				r=r_min+(j*r_step);
	
	//***************************************
	//initial condition
				mx=0;	my=0;	mz=1; ph=0;
//#######################################################################
				t=0;
				do{ //Here the cycle of time begins
					if((t*TP>=(t1-(TP/2.)))&&(t*TP<=(t2-(TP/2.)))){Signal=As;}else{Signal=0;}
            //Starting of Gauss-Legendre method
            // internal iteration for calculating k with fixed point method
                // 1. prepearing the initial approximations for k1 and k2
                // k1s  = f[x, y];
					mx1s=fmx(t, mx, my, mz, ph, G, r, alpha);
					my1s=fmy(t, mx, my, mz, ph, G, r, alpha);
					mz1s=fmz(t, mx, my, mz, ph, G, r, alpha);
					P1s=fph(t, mx, my, mz, ph, Signal);
                // k2s = f[x + h, y + h*k1s];
					mx2s=fmx(t+TP, mx+TP*mx1s, my+TP*my1s, mz+TP*mz1s, ph+TP*P1s, G, r, alpha);
					my2s=fmy(t+TP, mx+TP*mx1s, my+TP*my1s, mz+TP*mz1s, ph+TP*P1s, G, r, alpha);
					mz2s=fmz(t+TP, mx+TP*mx1s, my+TP*my1s, mz+TP*mz1s, ph+TP*P1s, G, r, alpha);
					P2s=fph(t+TP, mx+TP*mx1s, my+TP*my1s, mz+TP*mz1s, ph+TP*P1s, Signal);

            //cout<<"\ntime\t"<<time<<endl;
            //cout<<"k1s = "<<mx1s<<", "<<my1s<<", "<<mz1s<<", "<<P1s<<endl;
            //cout<<"k2s = "<<mx2s<<", "<<my2s<<", "<<mz2s<<", "<<P2s<<endl;

           /* fprintf(fdebug,"%.5f\t\n%.15f\t%.15f\t%.15f\t%.15f\t\n%.15f\t%.15f\t%.15f\t%.15f\t\n%.15f\t%.15f\t%.15f\t%.15f\t\n\n",
                     t*TP,
                     mx,my,mz,ph,
                     mx1s,my1s,mz1s,
                     P1s,mx2s,my2s,mz2s,P2s);*/

					s = 0;
					do{ //Here the cycle of fix point method for finding k1 and k2 begins
			    //cout<<"s = "<<s<<endl;
			    // k1s1 = f[x + c1 h, y + h*(a11*k1s + a12*k2s)];
						mx1s1=fmx(t + c1*TP, mx + TP*(a11*mx1s + a12*mx2s),
											my + TP*(a11*my1s + a12*my2s),
											mz + TP*(a11*mz1s + a12*mz2s),
											ph + TP*(a11*P1s + a12*P2s),G, r, alpha);
						my1s1=fmy(t + c1*TP, mx + TP*(a11*mx1s + a12*mx2s),
											my + TP*(a11*my1s + a12*my2s),
											mz + TP*(a11*mz1s + a12*mz2s),
											ph + TP*(a11*P1s + a12*P2s),G, r, alpha);
						mz1s1=fmz(t + c1*TP, mx + TP*(a11*mx1s + a12*mx2s),
											my + TP*(a11*my1s + a12*my2s),
											mz + TP*(a11*mz1s + a12*mz2s),
											ph + TP*(a11*P1s + a12*P2s),G, r, alpha);
						P1s1=fph(t + c1*TP,  mx + TP*(a11*mx1s + a12*mx2s),
											my + TP*(a11*my1s + a12*my2s),
											mz + TP*(a11*mz1s + a12*mz2s),
											ph + TP*(a11*P1s + a12*P2s), Signal);
                // k2s1 = f[x + c2 h, y + h*(a21*k1s + a22*k2s)];
						mx2s1=fmx(t + c2*TP, mx + TP*(a21*mx1s + a22*mx2s),
											my + TP*(a21*my1s + a22*my2s),
											mz + TP*(a21*mz1s + a22*mz2s),
											ph + TP*(a21*P1s + a22*P2s),G, r, alpha);
						my2s1=fmy(t + c2*TP, mx + TP*(a21*mx1s + a22*mx2s),
											my + TP*(a21*my1s + a22*my2s),
											mz + TP*(a21*mz1s + a22*mz2s),
											ph + TP*(a21*P1s + a22*P2s),G, r, alpha);
						mz2s1=fmz(t + c2*TP, mx + TP*(a21*mx1s + a22*mx2s),
											my + TP*(a21*my1s + a22*my2s),
											mz + TP*(a21*mz1s + a22*mz2s),
											ph + TP*(a21*P1s + a22*P2s),G, r, alpha);
						P2s1=fph(t + c2*TP,  mx + TP*(a21*mx1s + a22*mx2s),
											my + TP*(a21*my1s + a22*my2s),
											mz + TP*(a21*mz1s + a22*mz2s),
											ph + TP*(a21*P1s + a22*P2s), Signal);
                // k1s = k1s1; k2s = k2s1;
						mx1s = mx1s1; my1s = my1s1; mz1s = mz1s1; P1s = P1s1;
						mx2s = mx2s1; my2s = my2s1; mz2s = mz2s1; P2s = P2s1;

					}while(++s<=5);
			// y = y + 1/2 h*(k1s + k2s);
				mx=mx+(1/2.)*TP*(mx1s+mx2s);
				my=my+(1/2.)*TP*(my1s+my2s);
				mz=mz+(1/2.)*TP*(mz1s+mz2s);
				ph=ph+(1/2.)*TP*(P1s+P2s);
		//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
				m=sqrt(pow(mx,2)+pow(my,2)+pow(mz,2));

				Is=sin(ph-r*my);
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				time=t*TP;

            //cout<<"\ntime\t"<<time<<endl;

            /*fprintf(fIs,"%f\t%.15f\n",time,Is);
            fprintf(fsignal,"%f\t%.15f\n",time,Signal);
            fprintf(fphtime,"%f\t%.15f\n",time,ph);
            fprintf(fmxtime,"%f\t%.15f\n",time,mx);
            fprintf(fmytime,"%f\t%.15f\n",time,my);
            fprintf(fmztime,"%f\t%.15f\n",time,mz);
			fprintf(fmtime,"%f\t%.15f\n",time,m);
			fprintf(fmzmx,"%.15f\t%.15f\n",mx,mz);
			fprintf(fmzmy,"%.15f\t%.15f\n",my,mz);
			fprintf(fmymx,"%.15f\t%.15f\n",mx,my);*/
				}while(++t<=tN+1);
			
				#pragma omp critical
				{
					if(abs(mz+1)<=0.0001){
						point_arr_mz_p1[k_mz_p1].G=G;
						point_arr_mz_p1[k_mz_p1].r=r;
						point_arr_mz_p1[k_mz_p1].m=m;
						k_mz_p1++;
			//printf("%f %f %f\n",G,alpha,m);
					}
		
					if(abs(mz-1)<=0.0001){
						point_arr_mz_m1[k_mz_m1].G=G;
						point_arr_mz_m1[k_mz_m1].r=r;
						point_arr_mz_m1[k_mz_m1].m=m;
						k_mz_m1++;
			//printf("%f %f %f\n",G,alpha,m);
					}
		
		
					if((abs(mz)<=0.0001)&&(abs(my-1)<=0.0001)){
						point_arr_mz_my_m1[k_mz_my_m1].G=G;
						point_arr_mz_my_m1[k_mz_my_m1].r=r;
						point_arr_mz_my_m1[k_mz_my_m1].m=m;
						k_mz_my_m1++;
			//printf("%f %f %f\n",G,alpha,m);
					}
					iterations++;
					printf("%d of %d iterations calculated (thread: %d) \n",iterations,all_iter,rank);
				}
		
			}	
		
		}
	}
	stop=omp_get_wtime();
	FILE *po1, *po2, *po3;
	
	po1=fopen("points_Gr_mz_p1.dat","w");
	po2=fopen("points_Gr_mz_m1.dat","w");
	po3=fopen("points_Gr_mz_my_m1.dat","w");
	
	for(i=0;i<k_mz_p1;i++){
			fprintf(po1,"%f %f %.15f\n",point_arr_mz_p1[i].G,point_arr_mz_p1[i].r,point_arr_mz_p1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
	
	for(i=0;i<k_mz_m1;i++){
			fprintf(po2,"%f %f %.15f\n",point_arr_mz_m1[i].G,point_arr_mz_m1[i].r,point_arr_mz_m1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
	
	for(i=0;i<k_mz_my_m1;i++){
			fprintf(po3,"%f %f %.15f\n",point_arr_mz_my_m1[i].G,point_arr_mz_my_m1[i].r,point_arr_mz_my_m1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
			
	fclose(po1);
	fclose(po2);
	fclose(po3);
	
	printf("Done\n");
	printf("Calculation time: %f seconds \n",stop-start);
		/*if(abs(mz+1)<=0.0001){
			printf("%d \n%f \n%f\n",1,G,alpha);
		}
		else{
			printf("%d \n%f \n%f\n",0,G,alpha);

		}*/

	/*fclose(fIs);

    fclose(fsignal);
    fclose(fphtime);

	fclose(fmxtime);
	fclose(fmytime);
	fclose(fmztime);

	fclose(fmzmx);
	fclose(fmzmy);
	fclose(fmymx);
	fclose(fmtime);

	fclose(fdebug);*/

	return(0);
}

