//Spin orbital interaction Buzdin model 21-11-2016
//For the fixed value of Josephsn frequency
//Signal is added to the voltage as gauss distribution with integral

// Pavlina's changes:
//18.03.2018:
// - stop the while cycle of t with while(++t<=tN+1); - that way is adding also the last point
// - all physical and numerical parameters are taken above before the main
// - define functions fmx, fmy, fmy and check the correcteness of The first coefficients of Runge-Kutta method
//19.03.2018:
// - define function fph,
// - check the correcteness of The second, third and fourth coefficients of Runge-Kutta method
// - v2: clean the code
//22.03.2018:
// - adding h to the code (20min work)

#include<iostream>
#include<cmath>
#include<iomanip>
#include<stdio.h>
#include<stdlib.h>
#include <stddef.h>
#include <omp.h>

#include <cstdlib>
#include <cstdio>
#include <iostream>
#include <fstream>
#include <sstream>
#define MAXLEN 80

typedef struct point_s {
        double G;
        double r;
		double m;
} point;

struct sample_parameters
{
  char i_OmegaF[MAXLEN];
  char i_r_min[MAXLEN];
  char i_r_max[MAXLEN];
  char i_r_step[MAXLEN];
  char i_G_min[MAXLEN];
  char i_G_max[MAXLEN];
  char i_G_step[MAXLEN];
}
  sample_parameters;
  
 

/*
 * initialize data to default values
 */
void
init_parameters (struct sample_parameters * parms)
{
  strncpy (parms->i_OmegaF, "-25", MAXLEN);
  strncpy (parms->i_r_min, "-25", MAXLEN);
  strncpy (parms->i_r_max, "-25", MAXLEN);
  strncpy (parms->i_r_step, "-25", MAXLEN);
  strncpy (parms->i_G_min, "-25", MAXLEN);
  strncpy (parms->i_G_max, "-25", MAXLEN);
  strncpy (parms->i_G_step, "-25", MAXLEN);
 
}

/*
 * trim: get rid of trailing and leading whitespace...
 *       ...including the annoying "\n" from fgets()
 */
char *
trim (char * s)
{
  /* Initialize start, end pointers */
  char *s1 = s, *s2 = &s[strlen (s) - 1];

  /* Trim and delimit right side */
  while ( (isspace (*s2)) && (s2 >= s1) )
    s2--;
  *(s2+1) = '\0';

  /* Trim left side */
  while ( (isspace (*s1)) && (s1 < s2) )
    s1++;

  /* Copy finished string */
  strcpy (s, s1);
  return s;
}

/*
 * parse external parameters file
 *
 * NOTES:
 * - There are millions of ways to do this, depending on your
 *   specific needs.
 *
 * - In general:
 *   a) The client will know which parameters it's expecting
 *      (hence the "struct", with a specific set of parameters).
 *   b) The client should NOT know any specifics about the
 *      configuration file itself (for example, the client
 *      shouldn't know or care about it's name, its location,
 *      its format ... or whether or not the "configuration
 *      file" is even a file ... or a database ... or something
 *      else entirely).
 *   c) The client should initialize the parameters to reasonable
 *      defaults
 *   d) The client is responsible for validating whether the
 *      pararmeters are complete, or correct.
 */
void
parse_config (struct sample_parameters * parms, char * fl)
{
  char *s, buff[256];
  FILE *fp = fopen (fl, "r");
  if (fp == NULL)
  {
    return;
  }

  /* Read next line */
  while ((s = fgets (buff, sizeof buff, fp)) != NULL)
  {
    /* Skip blank lines and comments */
    if (buff[0] == '\n' || buff[0] == '#')
      continue;

    /* Parse name/value pair from line */
    char name[MAXLEN], value[MAXLEN];
    s = strtok (buff, "=");
    if (s==NULL)
      continue;
    else
      strncpy (name, s, MAXLEN);
    s = strtok (NULL, "=");
    if (s==NULL)
      continue;
    else
      strncpy (value, s, MAXLEN);
    trim (value);

    /* Copy into correct entry in parameters struct */
    if (strcmp(name, "OmegaF")==0)
      strncpy (parms->i_OmegaF, value, MAXLEN);
    else if (strcmp(name, "r_min")==0)
      strncpy (parms->i_r_min, value, MAXLEN);
    else if (strcmp(name, "r_max")==0)
      strncpy (parms->i_r_max, value, MAXLEN);
  else if (strcmp(name, "r_step")==0)
      strncpy (parms->i_r_step, value, MAXLEN);
  else if (strcmp(name, "G_min")==0)
      strncpy (parms->i_G_min, value, MAXLEN);
  else if (strcmp(name, "G_max")==0)
      strncpy (parms->i_G_max, value, MAXLEN);
  else if (strcmp(name, "G_step")==0)
      strncpy (parms->i_G_step, value, MAXLEN);
    else
      printf ("WARNING: %s/%s: Unknown name/value pair!\n",
        name, value);
  }

  /* Close file */
  fclose (fp);
}


//######################################################################################
// declearing all physical and numerical parameters
double
		Tmax=1000,//Time domain for one step of current
		TP=0.01,//Step of time
		Omega=0,//Josephson frequency
		OmegaF=1,
		
		
		alpha=0.5,//Dissipation parameter
		
		G_min=0,
		G_max=200,
		G_step=1,
		G,	//
		
		r,
		r_min=0,
		r_max=1,
		r_step=0.01,//
		
        //signal parameters
		As=1.5,//Амплитуда сигнала
		dt=6,//Среднеквадратичное отклонение
		t0=25,//Математическое ожидание
		V0=1, // = 1/w
		M=1;//Absolute value of the magnetic momentum

//######################################################################################
// declearing the functions of the mathematical model
double h_x (double t, double  mx, double my, double mz, double ph)
// hx
{
  double res;
    res=0;
  return  res;
}
double h_y (double t, double  mx, double my, double mz, double ph)
// hy
{
  double res;
    res=G*r*sin(ph-r*my);
  return  res;
}
double h_z (double t, double  mx, double my, double mz, double ph)
// hz
{
  double res;
    res=mz;
  return  res;
}
double fmx (double t, double  mx, double my, double mz, double ph)
// right-hand-side of dmx/dt = RHS equation
{
  double res, hx, hy, hz;
    hx = h_x(t,mx,my,mz,ph);
    hy = h_y(t,mx,my,mz,ph);
    hz = h_z(t,mx,my,mz,ph);
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				(my*hz - mz*hy)
                + alpha*(mx*(mx*hx+my*hy+mz*hz)-hx)
			);
  return  res;
}
double fmy (double t, double  mx, double my, double mz, double ph)
// right-hand-side of dmy/dt = RHS equation
{
  double res, hx, hy, hz;
    hx = h_x(t,mx,my,mz,ph);
    hy = h_y(t,mx,my,mz,ph);
    hz = h_z(t,mx,my,mz,ph);
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				(mz*hx-mx*hz)
				+alpha*(my*(mx*hx+my*hy+mz*hz)-hy)
			);
  return  res;
}

double fmz (double t, double  mx, double my, double mz, double ph)
// right-hand-side of dmz/dt = RHS equation
{
  double res, hx, hy, hz;
    hx = h_x(t,mx,my,mz,ph);
    hy = h_y(t,mx,my,mz,ph);
    hz = h_z(t,mx,my,mz,ph);
    res=-(OmegaF/(1+M*pow(alpha, 2)))*(
				mx*hy-my*hx+
				+alpha*(mz*(mx*hx+my*hy+mz*hz)-hz)
			);
  return  res;
}

double fph (double t, double  mx, double my, double mz, double ph, double Signal)
// right-hand-side of dph/dt = RHS equation
{
  double res;
    res=V0*(Signal-sin(ph-r*my));
  return  res;
}

int main()
	{
	using namespace std;
	
	FILE *stream;
	char mys[255];
	char str[50];
	
	struct sample_parameters parms;
	char * fl=argv[1];
	
	printf ("Initializing parameters to default values...\n");
	init_parameters (&parms);
	
	printf ("Reading config file...\n");
	parse_config (&parms,fl);
	
	OmegaF=atof(parms.i_OmegaF);

	r_min=atof(parms.i_r_min);
	r_max=atof(parms.i_r_max);
	r_step=atof(parms.i_r_step);
	
	G_min=atof(parms.i_G_min);
	G_max=atof(parms.i_G_max);
	G_step=atof(parms.i_G_step);
	
	int
		t,	//index of time
		tN;	//total number of time cycles
//######################################################################################
	double const
		pi=3.14159265358979323846;

	double
		//~~~~~~~~~~~~~~~~~~~
		Ej=1,//Josephson energy
		Es,//Superconducting part of Energy
		Signal,
		Is,
		time,//time value for writing
		m,//Absolute value of the magnetic momentum for writing
        t1,t2,
		ph,//phase difference of JJ
		P1,P2,P3,P4,
		//components of magnetic momentum
		mx,	 mx1,	mx2,	mx3,	mx4,
		my,	 my1,	my2,	my3,	my4,
		mz,	 mz1,	mz2,	mz3,	mz4;

		t1=t0-(dt/2.);
		t2=t0+(dt/2.);
//#######################################################################################

	//Creation of streams for the result's files
	/*FILE
		*fmxtime,	*fmytime,	*fmztime,   *fmtime,
		*fsignal,   *fphtime,
		*fmzmx,		*fmzmy,     *fmymx,
		*fes,       *fIs;
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	//Opening files
	fmxtime=fopen("mx_time.dat","w");
	fmytime=fopen("my_time.dat","w");
	fmztime=fopen("mz_time.dat","w");
	fsignal=fopen("Signal_time.dat","w");
	fphtime=fopen("ph_time.dat","w");
	fmtime=fopen("m_time.dat","w");
	fmzmx=fopen("mz_mx.dat","w");
	fmzmy=fopen("mz_my.dat","w");
	fmymx=fopen("my_mx.dat","w");
	fIs=fopen("Is_time.dat","w");*/
	//*****************************************************************
	tN=(Tmax/TP);//number of points in time
	//***************************************
	double start=0.0, stop=0.0;
	int rank, iterations=0.0;
	
	int i,j,k_mz_p1=0,k_mz_m1=0,k_mz_my_m1=0;
	int G_iter=(G_max-G_min)/G_step, r_iter=(r_max-r_min)/r_step;
	int all_iter=G_iter*r_iter;
	
	point point_arr_mz_p1[G_iter*r_iter],point_arr_mz_m1[G_iter*r_iter],point_arr_mz_my_m1[G_iter*r_iter];
	
	//begining of parallel region
	start=omp_get_wtime();
	#pragma omp parallel private(i,j,G,r,rank,mx, my, mz, ph, t, Signal, mx1, my1, mz1, mx2, my2, mz2, mx3, my3, mz3, mx4, my4, mz4, P1, P2, P3, P4, m, Is, time)
	{
		rank=omp_get_thread_num();
		#pragma omp for
		for(i=0;i<G_iter;i++){
			G=G_min+(i*G_step);
			for(j=0;j<r_iter;j++){
				r=r_min+(j*r_step);
	
	//initial condition
				mx=0;	my=0;	mz=1; ph=0;
//#######################################################################
				t=0;
				do{ //Here the cycle of time begins
					if((t*TP>(t1-(TP/2.)))&&(t*TP<(t2-(TP/2.)))){Signal=As;}else{Signal=0;}
			//Starting of Runge-Kutta method
		//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
			//The first coefficients of Runge-Kutta method

					mx1=TP*fmx(t, mx, my, mz, ph);
					my1=TP*fmy(t, mx, my, mz, ph);
					mz1=TP*fmz(t, mx, my, mz, ph);
					P1=TP*fph(t, mx, my, mz, ph, Signal);

			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			//The second coefficients of Runge-Kutta method

					mx2=TP*fmx(t+TP/2., mx+mx1/2., my+my1/2., mz+mz1/2., ph+P1/2.);
					my2=TP*fmy(t+TP/2., mx+mx1/2., my+my1/2., mz+mz1/2., ph+P1/2.);
					mz2=TP*fmz(t+TP/2., mx+mx1/2., my+my1/2., mz+mz1/2., ph+P1/2.);
					P2=TP*fph(t+TP/2., mx+mx1/2., my+my1/2., mz+mz1/2., ph+P1/2., Signal);

			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			//The third coefficients of Runge-Kutta method

					mx3=TP*fmx(t+TP/2., mx+mx2/2., my+my2/2., mz+mz2/2., ph+P2/2.);
					my3=TP*fmy(t+TP/2., mx+mx2/2., my+my2/2., mz+mz2/2., ph+P2/2.);
					mz3=TP*fmz(t+TP/2., mx+mx2/2., my+my2/2., mz+mz2/2., ph+P2/2.);
					P3=TP*fph(t+TP/2., mx+mx2/2., my+my2/2., mz+mz2/2., ph+P2/2., Signal);

			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			//The fourth coefficients of Runge-Kutta method

					mx4=TP*fmx(t+TP, mx+mx3, my+my3, mz+mz3, ph+P3);
					my4=TP*fmy(t+TP, mx+mx3, my+my3, mz+mz3, ph+P3);
					mz4=TP*fmz(t+TP, mx+mx3, my+my3, mz+mz3, ph+P3);
					P4=TP*fph(t+TP, mx+mx3, my+my3, mz+mz3, ph+P3, Signal);

			//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					mx=mx+(mx1+2*mx2+2*mx3+mx4)/6;
					my=my+(my1+2*my2+2*my3+my4)/6;
					mz=mz+(mz1+2*mz2+2*mz3+mz4)/6;

					ph=ph+(P1+2*P2+2*P3+P4)/6;
		//Ending of Runge-Kutta method
		//$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
					m=sqrt(pow(mx,2)+pow(my,2)+pow(mz,2));

					Is=sin(ph-r*my);
            //~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					time=t*TP;

            /*cout<<"\ntime\t"<<time<<endl;

            fprintf(fIs,"%f\t%.15f\n",time,Is);
            fprintf(fsignal,"%f\t%.15f\n",time,Signal);
            fprintf(fphtime,"%f\t%.15f\n",time,ph);
            fprintf(fmxtime,"%f\t%.15f\n",time,mx);
            fprintf(fmytime,"%f\t%.15f\n",time,my);
            fprintf(fmztime,"%f\t%.15f\n",time,mz);
			fprintf(fmtime,"%f\t%.15f\n",time,m);
			fprintf(fmzmx,"%.15f\t%.15f\n",mx,mz);
			fprintf(fmzmy,"%.15f\t%.15f\n",my,mz);
			fprintf(fmymx,"%.15f\t%.15f\n",mx,my);*/
				}while(++t<=tN+1);
			
			#pragma omp critical
			{
				if(abs(mz+1)<=0.0001){
					point_arr_mz_p1[k_mz_p1].G=G;
					point_arr_mz_p1[k_mz_p1].r=r;
					point_arr_mz_p1[k_mz_p1].m=m;
					k_mz_p1++;
			//printf("%f %f %f\n",G,alpha,m);
				}
		
				if(abs(mz-1)<=0.0001){
					point_arr_mz_m1[k_mz_m1].G=G;
					point_arr_mz_m1[k_mz_m1].r=r;
					point_arr_mz_m1[k_mz_m1].m=m;
					k_mz_m1++;
			//printf("%f %f %f\n",G,alpha,m);
				}
		
		
				if((abs(mz)<=0.0001)&&(abs(my-1)<=0.0001)){
					point_arr_mz_my_m1[k_mz_my_m1].G=G;
					point_arr_mz_my_m1[k_mz_my_m1].r=r;
					point_arr_mz_my_m1[k_mz_my_m1].m=m;
					k_mz_my_m1++;
			//printf("%f %f %f\n",G,alpha,m);
				}
				iterations++;
				printf("%d of %d iterations calculated (thread: %d) \n",iterations,all_iter,rank);
			}

			}
		
		}
	}
	
	FILE *po1, *po2, *po3;
	
	po1=fopen("points_Gr_mz_p1.dat","w");
	po2=fopen("points_Gr_mz_m1.dat","w");
	po3=fopen("points_Gr_mz_my_m1.dat","w");
	
	for(i=0;i<k_mz_p1;i++){
			fprintf(po1,"%f %f %.15f\n",point_arr_mz_p1[i].G,point_arr_mz_p1[i].r,point_arr_mz_p1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
	
	for(i=0;i<k_mz_m1;i++){
			fprintf(po2,"%f %f %.15f\n",point_arr_mz_m1[i].G,point_arr_mz_m1[i].r,point_arr_mz_m1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
	
	for(i=0;i<k_mz_my_m1;i++){
			fprintf(po3,"%f %f %.15f\n",point_arr_mz_my_m1[i].G,point_arr_mz_my_m1[i].r,point_arr_mz_my_m1[i].m);
			//printf("%f %f %f\n",point_arr[i].G,point_arr[i].alpha,point_arr[i].m);		
	}
			
	fclose(po1);
	fclose(po2);
	fclose(po3);
	
	printf("Done\n");
	/*fclose(fIs);

    fclose(fsignal);
    fclose(fphtime);

	fclose(fmxtime);
	fclose(fmytime);
	fclose(fmztime);

	fclose(fmzmx);
	fclose(fmzmy);
	fclose(fmymx);
	fclose(fmtime);*/

	return(0);
}

